const DATA_SOURCE_QUERY_OPERATORS = [
	'=',
	'!=',
	'>=',
	'<=',
	'>',
	'<',
	'LIKE',
	'ILIKE',
	'REGEXP'
] as const;

type DataSourceQueryOperator = typeof DATA_SOURCE_QUERY_OPERATORS[number];

interface DataSourceQueryExpression {
	lhs: string;
	operator: DataSourceQueryOperator;
	rhs: string | number | boolean | RegExp;
}

interface DataSourceQueryOrGrouping {
	type: 'or';
	expressions: DataSourceQueryExpression[];
}

interface DataSourceQueryAndGrouping {
	type: 'and';
	expressions: DataSourceQueryOrGrouping[];
}

export type DataSourceQuery = DataSourceQueryAndGrouping;

interface ProcessEntryBase {
	type: string;
}

const AVAILABLE_MATCH_TYPES = [
	'startsWith',
	'endsWith',
	'includes',
	'regexp',
] as const;

type MatchType = typeof AVAILABLE_MATCH_TYPES[number];

interface ProcessEntryString extends ProcessEntryBase {
	type: 'string';
	matchType?: MatchType;
	caseInsensitiveMatch?: boolean;
}

interface ProcessEntryNumber extends ProcessEntryBase {
	type: 'number';
	decimal?: boolean;
}

interface ProcessEntryBoolean extends ProcessEntryBase {
	type: 'boolean';
	truthyStrings?: string[];
}

export type ProcessEntry = ProcessEntryString | ProcessEntryNumber | ProcessEntryBoolean;

interface ConvertToDataSourceQueryCollectionOptions {
	processEntries: Record<string, ProcessEntry>;
}

const normalizeRhs = (lhs: string, rhs: string, processEntriesMap?: Record<string, ProcessEntry>) => {
	const coerceValues = processEntriesMap?.[lhs] ?? {
		type: 'string'
	};

	if (coerceValues?.type === 'number') {
		return {
			lhs,
			operator: '=',
			rhs: Number(rhs)
		} as DataSourceQueryExpression;
	}

	if (coerceValues?.type === 'boolean') {
		const truthyStrings = [
			...(coerceValues.truthyStrings ?? []).map((s) => s.trim().toLowerCase()),
			'true',
		];

		return {
			lhs,
			operator: '=',
			rhs: truthyStrings.includes(rhs)
		} as DataSourceQueryExpression;
	}

	if (coerceValues?.type === 'string') {
		switch (coerceValues?.matchType) {
			case 'startsWith': {
				return {
					lhs,
					operator: coerceValues.caseInsensitiveMatch ? 'ILIKE' : 'LIKE',
					rhs: `%${rhs}`,
				} as DataSourceQueryExpression;
			}
			case 'endsWith': {
				return {
					lhs,
					operator: coerceValues.caseInsensitiveMatch ? 'ILIKE' : 'LIKE',
					rhs: `${rhs}%`,
				} as DataSourceQueryExpression;
			}
			case 'includes': {
				return {
					lhs,
					operator: coerceValues.caseInsensitiveMatch ? 'ILIKE' : 'LIKE',
					rhs: `%${rhs}%`,
				} as DataSourceQueryExpression;
			}
			case 'regexp': {
				return {
					lhs,
					operator: 'REGEXP',
					rhs: new RegExp(rhs, coerceValues.caseInsensitiveMatch ? 'i' : ''),
				} as DataSourceQueryExpression;
			}
			default:
				break;
		}

		return {
			lhs,
			operator: '=',
			rhs,
		} as DataSourceQueryExpression;
	}

	const unknownCoerceValues = coerceValues as unknown as Record<string, string>;
	throw new Error(`Invalid coercion type: ${unknownCoerceValues.type}`);
	// this will be sent to the data source, e.g. the SQL query
	// we can also make this function act as a "sanitizer"
}

export const convertToDataSourceQuery = (
	q: URLSearchParams,
	options = {} as ConvertToDataSourceQueryCollectionOptions,
): DataSourceQueryAndGrouping => {
	return Array.from(q.entries()).reduce(
		(queries, [key, value]) => {
			const existingOr = queries.expressions
				.find((ex) => ex.expressions.some((ex2) => ex2.lhs === key)) ?? {
					type: 'or',
					expressions: [],
				};
			const existingLhs = existingOr.expressions.find((ex) => ex.lhs === key);
			const newExpression = normalizeRhs(key, value, options.processEntries);

			if (typeof existingLhs === 'undefined') {
				return {
					...queries,
					expressions: [
						...queries.expressions,
						{
							type: 'or',
							expressions: [
								newExpression,
							],
						},
					],
				};
			}

			return {
				...queries,
				expressions: queries.expressions.map((ex) => (
					ex.expressions.some((ex2) => ex2.lhs !== key)
						? ex
						: {
							...existingOr,
							expressions: [
								...(existingOr.expressions ?? []),
								newExpression,
							],
						}
				)),
			};
		},
		{
			type: 'and',
			expressions: [],
		} as DataSourceQueryAndGrouping
	)

	// all the queries are to be treated as "AND", as suggested by the & character to separate the query param entries
};

// TODO add more operators
