import {QueryMediaType, DataSourceQuery, DataSourceQueryExpression} from '../../common';

interface ProcessEntryBase {
	type: string;
}

const AVAILABLE_MATCH_TYPES = [
	'startsWith',
	'endsWith',
	'includes',
	'regexp',
] as const;

type MatchType = typeof AVAILABLE_MATCH_TYPES[number];

interface ProcessEntryString extends ProcessEntryBase {
	type: 'string';
	matchType?: MatchType;
	caseInsensitiveMatch?: boolean;
}

interface ProcessEntryNumber extends ProcessEntryBase {
	type: 'number';
	decimal?: boolean;
}

interface ProcessEntryBoolean extends ProcessEntryBase {
	type: 'boolean';
	truthyStrings?: string[];
}

export type ProcessEntry = ProcessEntryString | ProcessEntryNumber | ProcessEntryBoolean;

export const name = 'application/x-www-form-urlencoded' as const;

const normalizeRhs = (lhs: string, rhs: string, processEntriesMap?: Record<string, ProcessEntry>) => {
	const coerceValues = processEntriesMap?.[lhs] ?? {
		type: 'string'
	};

	if (coerceValues?.type === 'number') {
		return {
			lhs,
			operator: '=',
			rhs: Number(rhs)
		} as DataSourceQueryExpression;
	}

	if (coerceValues?.type === 'boolean') {
		const truthyStrings = [
			...(coerceValues.truthyStrings ?? []).map((s) => s.trim().toLowerCase()),
			'true',
		];

		return {
			lhs,
			operator: '=',
			rhs: truthyStrings.includes(rhs)
		} as DataSourceQueryExpression;
	}

	if (coerceValues?.type === 'string') {
		switch (coerceValues?.matchType) {
			case 'startsWith': {
				return {
					lhs,
					operator: coerceValues.caseInsensitiveMatch ? 'ILIKE' : 'LIKE',
					rhs: `%${rhs}`,
				} as DataSourceQueryExpression;
			}
			case 'endsWith': {
				return {
					lhs,
					operator: coerceValues.caseInsensitiveMatch ? 'ILIKE' : 'LIKE',
					rhs: `${rhs}%`,
				} as DataSourceQueryExpression;
			}
			case 'includes': {
				return {
					lhs,
					operator: coerceValues.caseInsensitiveMatch ? 'ILIKE' : 'LIKE',
					rhs: `%${rhs}%`,
				} as DataSourceQueryExpression;
			}
			case 'regexp': {
				return {
					lhs,
					operator: 'REGEXP',
					rhs: new RegExp(rhs, coerceValues.caseInsensitiveMatch ? 'i' : ''),
				} as DataSourceQueryExpression;
			}
			default:
				break;
		}

		return {
			lhs,
			operator: '=',
			rhs,
		} as DataSourceQueryExpression;
	}

	const unknownCoerceValues = coerceValues as unknown as Record<string, string>;
	throw new Error(`Invalid coercion type: ${unknownCoerceValues.type}`);
	// this will be sent to the data source, e.g. the SQL query
	// we can also make this function act as a "sanitizer"
}

interface SerializeOptions {}

interface DeserializeOptions {
	processEntries?: Record<string, ProcessEntry>;
}

export const deserialize: QueryMediaType<typeof name, SerializeOptions, DeserializeOptions>['deserialize'] = (
	s: string,
	options = {} as DeserializeOptions
) => {
	const q = new URLSearchParams(s);
	return Array.from(q.entries()).reduce(
		(queries, [key, value]) => {
			const existingOr = queries.expressions
				.find((ex) => ex.expressions.some((ex2) => ex2.lhs === key)) ?? {
				type: 'or',
				expressions: [],
			};
			const existingLhs = existingOr.expressions.find((ex) => ex.lhs === key);
			const newExpression = normalizeRhs(key, value, options.processEntries);

			if (typeof existingLhs === 'undefined') {
				return {
					...queries,
					expressions: [
						...queries.expressions,
						{
							type: 'or',
							expressions: [
								newExpression,
							],
						},
					],
				};
			}

			return {
				...queries,
				expressions: queries.expressions.map((ex) => (
					ex.expressions.some((ex2) => ex2.lhs !== key)
						? ex
						: {
							...existingOr,
							expressions: [
								...(existingOr.expressions ?? []),
								newExpression,
							],
						}
				)),
			};
		},
		{
			type: 'and',
			expressions: [],
		} as DataSourceQuery
	)
};

export const serialize: QueryMediaType<typeof name, SerializeOptions, DeserializeOptions>['serialize'] = (q: DataSourceQuery) => (
	new URLSearchParams(
		q.expressions.map(
			(ex) => ex.expressions.map((ex2) => {
				if (ex2.rhs instanceof RegExp) {
					if (ex2.operator !== 'REGEXP') {
						throw new Error(`Invalid rhs given for operator: ${ex2.lhs} ${ex2.operator} <rhs>`);
					}

					return [ex2.lhs, ex2.rhs.toString()];
				}

				switch (typeof ex2.rhs) {
					case 'string': {
						switch (ex2.operator) {
							case 'ILIKE':
							case 'LIKE':
								return [ex2.lhs, ex2.rhs.replace(/^%+/, '').replace(/%+$/, '')];
							case '=':
								return [ex2.lhs, ex2.rhs];
							default:
								break;
						}
						throw new Error(`Invalid operator given for lhs: ${ex2.lhs} <op> ${ex2.rhs}`);
					}
					case 'number': {
						return [ex2.lhs, ex2.rhs.toString()];
					}
					case 'boolean': {
						return [ex2.lhs, ex2.rhs ? 'true' : 'false'];
					}
					default:
						break;
				}

				throw new Error(`Unknown type for rhs: ${ex2.lhs} ${ex2.operator} <rhs>`);
			})
		)
			.flat()
	)
		.toString()
);
